#include "stdafx.h"
#include "PatternFile.hpp"
#include "MD5.hpp"
#include "Utility.hpp"
#include "PtTrackIndexes.hpp"

namespace Lmt {
namespace Max2Bms {
	PatternFile::PatternFile( void ) {
		m_ppt = NULL;
		m_hash = L"";
	}

	PatternFile::~PatternFile( void ) {
		this->Clear();
	}

	int PatternFile::GetSoundEntryCount( void ) const {
		return static_cast<int>( m_sentries.size() );
	}

	int PatternFile::GetTrackCount( void ) const {
		return static_cast<int>( m_tracks.size() );
	}

	int PatternFile::GetTotalNotes( void ) const {
		static const int trkIndexes[] = {
			P1Line1TrackIndex, P1Line2TrackIndex, P1Line3TrackIndex, P1Line4TrackIndex, 
			P1Line5TrackIndex, P1ScratchTrackIndex, P1PedalTrackIndex, -1
		};

		int totalNotes = 0;
		int elemCount;
		Track trk;
		TrackElement elem;

		for (int i = 0; trkIndexes[i] != -1; i++) {
			if (this->GetTrack( trkIndexes[i], trk ) == false) {
				continue;
			}

			elemCount = trk.GetElementCount();
			for (int j = 0; j < elemCount; j++) {
				trk.GetElement( j, elem );
				if (elem.GetType() == PlaySoundTrackElement) {
					totalNotes++;
				}
			}
		}

		return totalNotes;
	}

	wstring PatternFile::GetHash( void ) const {
		return m_hash;
	}

	bool PatternFile::GetSoundEntry( int index, SoundEntry &sentry ) const {
		SoundEntry sentryTemp;

		if ((index < 0) || (index >= this->GetSoundEntryCount())) {
			return false;
		}

		if (sentryTemp.Initialize( m_sentries[index] ) == false) {
			return false;
		}

		sentry = sentryTemp;
		return true;
	}

	bool PatternFile::GetTrack( int index, Track &track ) const {
		Track trackTemp;

		if ((index < 0) || (index >= this->GetTrackCount())) {
			return false;
		}

		if (trackTemp.Initialize( m_tracks[index] ) == false) {
			return false;
		}

		track = trackTemp;
		return true;
	}

	PtLoadResult PatternFile::Load( const wstring &path ) {
		MD5 md5;
		PtFormat::Header *pheader;
		PtFormat::TrackHeader *ptheader;
		SoundEntry sentryTemp;
		Track trackTemp;
		SoundEntryArray sentries;
		TrackArray tracks;
		fstream fs;
		unsigned char digest[16];
		unsigned char *ppt = NULL;
		unsigned char *ptemp;
		unsigned int fileSize;
		unsigned int trackSize;
		wstring hash;
		int i;
		PtLoadResult result = PtLoadOk;

		fs.open( Utility::WStringToString( path ).c_str(), ios::in | ios::binary );
		if (fs.is_open() == false) {
			result = PtLoadFailedToOpen;
			goto quit;
		}

		fs.seekg( 0, ios::end );
		fileSize = fs.tellg();
		fs.seekg( 0, ios::beg );

		ppt = new unsigned char [fileSize];
		if (ppt == NULL) {
			result = PtLoadFailedToOpen;
			goto quit;
		}

		fs.read( (char*)ppt, fileSize );
		if (fs.fail() == true) {
			result = PtLoadIOError;
			goto quit;
		}
		if (fs.gcount() != fileSize) {
			result = PtLoadNotPtFormat;
			goto quit;
		}

		pheader = reinterpret_cast<PtFormat::Header*>( ppt );
		if (memcmp( pheader->signature, "PTFF", 4 ) != 0) {
			result = PtLoadNotPtFormat;
			goto quit;
		}

		for (i = 0; i < static_cast<int>( pheader->numberOfSound ); i++) {
			ptemp = ppt + sizeof(*pheader) + (i * sizeof(PtFormat::SoundDef));

			if (sentryTemp.Initialize( ptemp ) == false) {
				result = PtLoadBadSoundEntry;
				goto quit;
			}

			sentries.insert( sentries.end(), ptemp );
		}

		ptemp = ppt + sizeof(*pheader) + (pheader->numberOfSound * sizeof(PtFormat::SoundDef));
		for (i = 0; i < static_cast<int>( pheader->numberOfEztr ); i++) {
			tracks.insert( tracks.end(), ptemp );

			if (trackTemp.Initialize( ptemp ) == false) {
				return PtLoadBadTrack;
			}

			ptheader = reinterpret_cast<PtFormat::TrackHeader*>( ptemp );
			ptemp += (sizeof(*ptheader) + ptheader->dataLength);
		}

		ptemp = ppt + sizeof(*pheader) + (pheader->numberOfSound * sizeof(PtFormat::SoundDef));
		trackSize = fileSize - (ptemp - ppt);
		md5.Start();
		md5.Update( ptemp, trackSize );
		md5.Finish( digest );
		hash = MD5::ToString( digest );

		m_ppt = ppt;
		m_sentries.swap( sentries );
		m_tracks.swap( tracks );
		m_hash.swap( hash );

	quit:
		if ((result != PtLoadOk) && (ppt != NULL)) {
			delete [] ppt;
		}
		return result;
	}

	void PatternFile::Clear( void ) {
		if (m_ppt != NULL) {
			delete [] m_ppt;
			m_sentries.clear();
			m_tracks.clear();
			m_hash = L"";
		}
	}
}
}
